//cell.cc: Implementation file for useful cells for CROWBAR
//logic gates, latches

#include "signal.h"
#include "primitives.h"
#include "custom.h"
#include "cell.h"


//Combinational logic
////////////////////////////////////////////////////////////////

signal_type TMUX4(signal_type* inputs, signal_type* selects)
{
  //4-way mux with three-state inverters (for CLB)
  //Two level, first bit selects first level

  signal_type first_level_select;
  signal_type first_level_select_not;
  signal_type first_level_output1;
  signal_type first_level_output2;

  signal_type second_level_select;
  signal_type second_level_select_not;
  signal_type second_level_output;

  //decode select lines
  first_level_select = selects[0];
  first_level_select_not = INV(first_level_select);
  second_level_select = selects[1];
  second_level_select_not = INV(second_level_select);

  //do first level
  first_level_output1 = Wired_OR2(TINV(inputs[0], first_level_select_not),
				  TINV(inputs[1], first_level_select) );
  first_level_output2 = Wired_OR2(TINV(inputs[2], first_level_select_not),
				  TINV(inputs[3], first_level_select) );

  //do second level
  second_level_output = Wired_OR2(TINV(first_level_output1, 
				       second_level_select_not),
				  TINV(first_level_output2, 
				       second_level_select) );
  //done
  return second_level_output;
} //TMUX4


signal_type TMUX8(signal_type* inputs, signal_type* selects)
{
  //for CSB
  //Three level, first bit selects first level
  signal_type first_level_select;
  signal_type first_level_select_not;
  signal_type first_level_output1;
  signal_type first_level_output2;
  signal_type first_level_output3;
  signal_type first_level_output4;

  signal_type second_level_select;
  signal_type second_level_select_not;
  signal_type second_level_output1;
  signal_type second_level_output2;

  signal_type third_level_select;
  signal_type third_level_select_not;
  signal_type third_level_output;

  //decode select lines
  first_level_select = selects[0];
  first_level_select_not = INV(first_level_select);
  second_level_select = selects[1];
  second_level_select_not = INV(second_level_select);
  third_level_select = selects[2];
  third_level_select_not = INV(third_level_select);

  //do first level
  first_level_output1 = Wired_OR2(TINV(inputs[0], first_level_select_not),
				  TINV(inputs[1], first_level_select) );
  first_level_output2 = Wired_OR2(TINV(inputs[2], first_level_select_not),
				  TINV(inputs[3], first_level_select) );
  first_level_output3 = Wired_OR2(TINV(inputs[4], first_level_select_not),
				  TINV(inputs[5], first_level_select) );
  first_level_output4 = Wired_OR2(TINV(inputs[6], first_level_select_not),
				  TINV(inputs[7], first_level_select) );

  //do second level
  second_level_output1 = Wired_OR2(TINV(first_level_output1, 
					second_level_select_not),
				   TINV(first_level_output2, 
					second_level_select) );
  second_level_output2 = Wired_OR2(TINV(first_level_output3, 
					second_level_select_not),
				   TINV(first_level_output4, 
					second_level_select) );
  //do third level
  third_level_output = Wired_OR2(TINV(second_level_output1, 
				      third_level_select_not),
				 TINV(second_level_output2, 
				      third_level_select) );

  //done
  return INV(third_level_output); //even number of inversions
} //TMUX8


signal_type GMUX2(signal_type inp0, signal_type inp1, signal_type select1)
{
  //2-way mux with gates (scan-mux, config-mux)
  //three NAND gates

  signal_type first_level_output1;
  signal_type first_level_output2;
  signal_type output;

  first_level_output1 = NAND2(select1, inp1);
  first_level_output2 = NAND2(INV(select1), inp0);
  output = NAND2(first_level_output1,first_level_output2);

  return output;
} //GMUX2


signal_type Pseudo_NMOS_MUX8(signal_type* inputs, signal_type* selects)
{
  //8-way mux used for look-up tables
  //stacked AND of nfets
  //P-NMOS OR of AND-ed products

  signal_type products[8];
  signal_type select1 = selects[0];
  signal_type select2 = selects[1];
  signal_type select3 = selects[2];
  signal_type select1_not = INV(select1);
  signal_type select2_not = INV(select2);
  signal_type select3_not = INV(select3);
  signal_type output;

  signal_type AND_inputs0[] = {select1_not, select2_not, select3_not, inputs[0]};
  signal_type AND_inputs1[] = {select1, select2_not, select3_not, inputs[1]};
  signal_type AND_inputs2[] = {select1_not, select2, select3_not, inputs[2]};
  signal_type AND_inputs3[] = {select1, select2, select3_not, inputs[3]};
  signal_type AND_inputs4[] = {select1_not, select2_not, select3, inputs[4]};
  signal_type AND_inputs5[] = {select1, select2_not, select3, inputs[5]};
  signal_type AND_inputs6[] = {select1_not, select2, select3, inputs[6]};
  signal_type AND_inputs7[] = {select1, select2, select3, inputs[7]};

  products[0] = ANDN(AND_inputs0, 4);
  products[1] = ANDN(AND_inputs1, 4);
  products[2] = ANDN(AND_inputs2, 4);
  products[3] = ANDN(AND_inputs3, 4);
  products[4] = ANDN(AND_inputs4, 4);
  products[5] = ANDN(AND_inputs5, 4);
  products[6] = ANDN(AND_inputs6, 4);
  products[7] = ANDN(AND_inputs7, 4);

  output = ORN(products, 8);
  return output;
} //Pseudo


//Latches: simple memory elements (conditional, resettable)
////////////////////////////////////////////////////////////////

conditional_latch::conditional_latch()
{
} //constructor


conditional_latch::~conditional_latch()
{
} //destructor


void conditional_latch::Update(signal_type clock, signal_type input, signal_type control)
{
  switch(control)
    {
    case True:
      switch(clock)
	{
	case True:
	  latch_state = input;
	  break;
	case False:
	  break;
	default:
	  latch_state = Unknown;
	} //switch
      break;
    case False: //do nothing
      break; 
    default: //may be a problem
      switch(clock)
	{
	case True:
	  latch_state = Unknown;
	  break;
	case False:
	  break;
	default:
	  latch_state = Unknown;
	} //switch
    } //switch
} //Update
		

resettable_latch::resettable_latch()
{
} //constructor


resettable_latch::~resettable_latch()
{
} //destructor


void resettable_latch::Update(signal_type clock, signal_type input, signal_type reset)
{
  switch(reset)
    {
    case True:
      switch(clock)
	{
	case True:
	  latch_state = Unknown;
	  break;
	case False:
	  latch_state = False;
	  break;
	default:
	  latch_state = Unknown;
	} //switch
      break;
    case False:
      switch(clock)
	{
	case True:
	  latch_state = input;
	  break;
	case False:
	  break;
	default:
	  latch_state = Unknown;
	} //switch
      break; 
    default: //problem
      latch_state = Unknown;
    } //switch
} //Update

//Flops: compound memory elements
////////////////////////////////////////////////////////////////

config_flop::config_flop()
{
  Clear();
} //constructor


config_flop::~config_flop()
{
} //destructor


void config_flop::Update(signal_type clk_a, signal_type clk_b,
			 signal_type enconf, signal_type input)
{
  b_latch.Update(clk_b, a_latch.Get_Output(), enconf);
  a_latch.Update(clk_a, input);
} //Update


signal_type config_flop::Get_Output()
{
  return b_latch.Get_Output();
} //Get_Output


void config_flop::Clear()
{
  a_latch.Clear();
  b_latch.Clear();
} //Clear


resettable_flop::resettable_flop()
{
  Clear();
} //constructor


resettable_flop::~resettable_flop()
{
} //destructor


void resettable_flop::Update(signal_type clk_a, signal_type clk_b,
			     signal_type reset, signal_type input)
{
  b_latch.Update(clk_b, a_latch.Get_Output(), reset);
  a_latch.Update(clk_a, input);
} //Update


signal_type resettable_flop::Get_Output()
{
  return b_latch.Get_Output();
} //Get_Output


void resettable_flop::Clear()
{
  a_latch.Clear();
  b_latch.Clear();
} //Clear


//Compound/Combined structures (now to do some real damage)
///////////////////////////////////////////////////////////////


LUT8::LUT8()
{
  Clear();
} //constructor


LUT8::~LUT8()
{
} //destructor


void LUT8::Update(signal_type clk_a, signal_type clk_b,
		  signal_type en_conf, signal_type shift_in)
{
  //updates shift register
  signal_type temp;
  temp = shift_in;
  for(unsigned int i = 0; i<LUT8_Config_Bits; i++)
    {
      LUT_entries[i].Update(clk_a, clk_b, en_conf, temp);
      temp = LUT_entries[i].Get_Output();
    } //for
} //Update


signal_type LUT8::Get_Output(signal_type* index)
{
  signal_type LUT_values[LUT8_Config_Bits];
  for(unsigned int i = 0; i<LUT8_Config_Bits; i++)
    {
      LUT_values[i] = LUT_entries[i].Get_Output();
    } //for
  return Pseudo_NMOS_MUX8(LUT_values, index);
} //Get_Output


signal_type LUT8::Get_Shift_Output()
{
  return LUT_entries[LUT8_Config_Bits-1].Get_Output();
} //Get_Shift_Output


signal_type LUT8::Get_Config_Bit(unsigned int index)
{
  if(index < LUT8_Config_Bits)
    {
      return LUT_entries[index].Get_Output();
    } //if
  else
    return Unknown;
} //Get_Config_Bit


void LUT8::Clear()
{
  for(unsigned int i = 0; i<LUT8_Config_Bits; i++)
    {
      LUT_entries[i].Clear();
    } //for
} //Clear


configurable_mux4::configurable_mux4()
{
  Clear();
} //constructor


configurable_mux4::~configurable_mux4()
{
} //destructor


void configurable_mux4::Update(signal_type clk_a, signal_type clk_b, 
			       signal_type en_conf, signal_type shift_in)
{
  //updates shift register
  signal_type temp;
  temp = shift_in;
  for(unsigned int i = 0; i<Configurable_Mux4_Config_Bits; i++)
    {
      config[i].Update(clk_a, clk_b, en_conf, temp);
      temp = config[i].Get_Output();
    } //for
} //Update


signal_type configurable_mux4::Get_Output(signal_type* inputs)
{
  signal_type selects[Configurable_Mux4_Config_Bits];
  for(unsigned int i = 0; i<Configurable_Mux4_Config_Bits; i++)
    {
      selects[i] = config[i].Get_Output();
    } //for
  return TMUX4(inputs, selects);
} //Get_Output


signal_type configurable_mux4::Get_Shift_Output()
{
  return config[Configurable_Mux4_Config_Bits-1].Get_Output(); //last bit
} //Get_Shift_Output


signal_type configurable_mux4::Get_Config_Bit(unsigned int index)
{
  if(index < Configurable_Mux4_Config_Bits)
    return config[index].Get_Output();
  else
    return Unknown;
} //Get_Config_Bit


void configurable_mux4::Clear()
{
  for(unsigned int i = 0; i<Configurable_Mux4_Config_Bits; i++)
    {
      config[i].Clear();
    } //for
} //Clear


configurable_mux8::configurable_mux8()
{
  Clear();
} //constructor


configurable_mux8::~configurable_mux8()
{
} //destructor


void configurable_mux8::Update(signal_type clk_a, signal_type clk_b, 
			       signal_type en_conf, signal_type shift_in)
{
  //updates shift register
  signal_type temp;
  temp = shift_in;
  for(unsigned int i = 0; i<Configurable_Mux8_Config_Bits; i++)
    {
      config[i].Update(clk_a, clk_b, en_conf, temp);
      temp = config[i].Get_Output();
    } //for
} //Update


signal_type configurable_mux8::Get_Output(signal_type* inputs)
{
  signal_type selects[Configurable_Mux8_Config_Bits];
  for(unsigned int i = 0; i<Configurable_Mux8_Config_Bits; i++)
    {
      selects[i] = config[i].Get_Output();
    } //for
  return TMUX8(inputs, selects);
} //Get_Output


signal_type configurable_mux8::Get_Shift_Output()
{
  return config[Configurable_Mux8_Config_Bits-1].Get_Output(); //last bit
} //Get_Shift_Output


signal_type configurable_mux8::Get_Config_Bit(unsigned int index)
{
  if(index < Configurable_Mux8_Config_Bits)
    return config[index].Get_Output();
  else
    return Unknown;
} //Get_Config_Bit


void configurable_mux8::Clear()
{
  for(unsigned int i = 0; i<Configurable_Mux8_Config_Bits; i++)
    {
      config[i].Clear();
    } //for
} //Clear


receiver::receiver()
{
  Clear();
} //constructor


receiver::~receiver()
{
} //destructor


void receiver::Update(signal_type clk_a, signal_type clk_b,
		      signal_type en_conf, signal_type shift_in)
{
  //updates shift register
  signal_type temp;
  temp = shift_in;
  for(unsigned int i = 0; i<Receiver_Config_Bits; i++)
    {
      config[i].Update(clk_a, clk_b, en_conf, temp);
      temp = config[i].Get_Output();
    } //for

  //update staging latches (when done configuring)
  signal_type fire_staging_latches = INV(en_conf);
  for(unsigned int i = 0; i<Receiver_Config_Bits; i++)
    {
      staging_latches[i].Update(fire_staging_latches, config[i].Get_Output());
    } //for
} //Update


signal_type receiver::Get_Output(optical_type optical_input)
{
  //Generate reference based on output of staging latches
  //First bit is most significant
  signal_type staging_latch_outputs[Receiver_Config_Bits];
  optical_type reference;

  for(unsigned int i = 0; i<Receiver_Config_Bits; i++)
    {
      staging_latch_outputs[(Receiver_Config_Bits-1)-i] = staging_latches[i].Get_Output();
    } //for
  reference = Voltage_Reference(staging_latch_outputs);

  //Compare input to reference
  return Diff_amp(optical_input, reference);
} //Get_Output


signal_type receiver::Get_Shift_Output()
{
  return config[Receiver_Config_Bits-1].Get_Output(); //last bit
} //Get_Shift_Output


signal_type receiver::Get_Config_Bit(unsigned int index)
{
  if(index < Receiver_Config_Bits)
    return config[index].Get_Output();
  else
    return Unknown;
} //Get_Config_Bit

void receiver::Clear()
{
  for(unsigned int i = 0; i<Receiver_Config_Bits; i++)
    {
      config[i].Clear();
    } //for
} //Clear


///////////////////////////////////////////////////////////////
//End of implementation file


